import sys,os.path as path
sys.path.append(path.dirname(path.dirname(path.dirname(path.abspath(__file__)))))

"""
Convert WX to UTF

To run:
python tagger/utils/convert_encoding.py -ie wx -oe utf -d ssf -f data/train/hin/train.wx.ssf -o outputs/train.utf.ssf -l hin

-ie - Input encoding
-oe - Output encoding
-d -  Data format (ssf)
-i -  Input file
-o -  Output file
"""

import re, os
import codecs
import logging
from wxconv import WXC
from tagger.utils.writer import write_to_file
import argparse
logger = logging.getLogger(__name__)

def convert_encoding(filename, text_type, language, in_enc, out_enc):

	with codecs.open(filename, 'r', encoding='utf-8') as fp:
		logger.info('Loading text_type: %s data' % (text_type))
		output_data = ""
		if text_type == "ssf":
			converter = WXC(order='%s2%s' % (in_enc, out_enc), lang=language)
			for line in fp:
				line = line.strip()
				ds = line.split()
				#print("Line:--", ds)
				if line == "":
					output_data += u"\n"
				elif line[0] == "<":
					output_data += u"%s\n" % (line)
				elif ds[0] == "))":
					output_data += u"\t%s\n" % (line)
				elif ds[0] == "0" or ds[1] == "((":
					output_data += u"%s\n" % (u"\t".join(ds))
				elif ds[1] != "":
					#print("check", ds)
					word, tag = ds[1], ds[2]
					word_con = converter.convert(word)
					output_data += u"%s\t%s\t%s\n" % (ds[0], word_con, tag)
				else:
					pass
			return output_data
		else:
			if text_type == "tnt":
				converter = WXC(order='%s2%s' % (in_enc, out_enc), lang=language, format_='tnt')
			elif text_type == "text":
				converter = WXC(order='%s2%s' % (in_enc, out_enc), lang=language)
			elif text_type == "conll":
				converter = WXC(order='%s2%s' % (in_enc, out_enc), lang=language, format_='conll')
			else:
				raise Exception("Unknown Format %s" % text_type)
				pass
       		text = fp.read()
        	output_data = converter.convert(text)
       		return output_data

def get_args():
    ''' This function parses and return arguments passed in'''
    parser = argparse.ArgumentParser(description='Encoding Converter for Indian Languages')
    parser.add_argument("-l", "--language", dest="language", type=str, metavar='<str>', required=True,
                         help="Language of the dataset: tel (telugu), hin (hindi), tam (tamil), kan (kannada), pun (pubjabi)")
    parser.add_argument("-ie", "--in_enc", dest="in_enc", type=str, metavar='<str>', required=True,
                        help="Input Encoding of the data (utf, wx)")
    parser.add_argument("-oe", "--out_enc", dest="out_enc", type=str, metavar='<str>', required=True,
                        help="Input Encoding of the data (utf, wx)")
    parser.add_argument("-d", "--data_format", dest="data_format", type=str, metavar='<str>', required=True,
                        help="Data format (ssf, tnt, txt)")
    parser.add_argument("-i", "--input_path", dest="file_path", type=str, metavar='<str>', required=True,
                        help="File to be converted file path ex: data/test/telugu/test.wx.ssf")
    parser.add_argument("-o", "--output_path", dest="output_path", type=str, metavar='<str>', required=True,
                         help="The path to the output file ex: outputs/test.utf.ssf")

    return parser.parse_args()


if __name__ == "__main__":

	args = get_args()

	output_data=convert_encoding(args.input_path, args.data_format, args.language, args.in_enc, args.out_enc)
	write_to_file(output_data, args.output_path)













'''
Pipeline code for Indic Tagger

Example:
python pipeline.py -p train -o outputs -l tel -t chunk -m crf -i data/test/tel/test.utf.conll.chunk -e utf -f conll

    -p, --pipeline_type - train, test, predict
    -l, --language      - te, hi, ta, pu, mr, be, ur, ka, ml
    -t, --tag_type      - pos, chunk
    -m, --model_type    - crf, hmm, cnn, lstm
    -f, --data_format   - ssf, tnt, text
    -e, --encoding      - utf8, wx (default: utf8)
    -i, --input_file    - path to the test data file
    -o, --output_file   - path to the output file
    -s, --sent_split    - split the sentences in the test data (default: True)
'''

from polyglot_tokenizer import Tokenizer
from hindiwsd.CRF import CRF
from sklearn.model_selection import train_test_split
from time import time
import numpy as np
import pickle
import logging
import argparse
import hindiwsd.writer as data_writer
import hindiwsd.generate_features as generate_features
import hindiwsd.data_reader as data_reader
#import hindiwsd.lstmcrf
from hindiwsd.utils import load_data_and_labels
from hindiwsd.wrapper import Sequence
import sys
import os.path as path
import os
sys.path.append(path.dirname(path.abspath(__file__)))


logger = logging.getLogger(__name__)


def get_args():
    ''' This function parses and return arguments passed in'''

    parser = argparse.ArgumentParser(description='Scorer pipeline')
    parser.add_argument("-p", '--pipeline_type', type=str, required=True,
                        help='Pipeline Type (train, test, predict)')
    parser.add_argument("-l", "--language", dest="language", type=str, metavar='<str>', required=True,
                        help="Language of the dataset: te (telugu), hi (hindi), ta (tamil), ka (kannada), pu (pubjabi), mr (Marathi), be (Bengali), ur (Urdu), ml (Malayalam)")
    parser.add_argument("-t", "--tag_type", dest="tag_type", type=str, metavar='<str>', required=True,
                        help="Tag type: pos, chunk, parse, NER")
    parser.add_argument("-m", "--model_type", dest="model_type", type=str, metavar='<str>', required=True,
                        help="Model type (crf|hmm|cnn|lstm:) (default=crf)")
    parser.add_argument("-e", "--encoding", dest="encoding", type=str, metavar='<str>', required=False,
                        help="Encoding of the data (utf8, wx)",
                        default="utf8")
    parser.add_argument("-f", "--data_format", dest="data_format", type=str, metavar='<str>', required=True,
                        help="Data format (ssf, tnt, txt)")
    parser.add_argument("-i", "--input_file", dest="test_data", type=str, metavar='<str>', required=False,
                        help="Test data path ex: data/test/te/test.txt")
    parser.add_argument("-s", "--sent_split", dest="sent_split", type=str, metavar='<str>', required=False,
                        help="Sentence Split ex: True or False",
                        default=True)
    parser.add_argument("-o", "--output_file", dest="output_path", type=str, metavar='<str>',
                        help="The path to the output file",
                        default=path.join(path.dirname(path.abspath(__file__)), "outputs", "output_file"))
    return parser.parse_args()


def pipeline():
    curr_dir = path.dirname(path.abspath(__file__))
    args = get_args()

    output_dir = path.join(path.dirname(path.abspath(__file__)), "outputs")
    if not os.path.exists(output_dir):
        os.makedirs(output_dir)

    data_writer.set_logger(args.model_type, output_dir)

    if True:
        model_path = "%s/models/%s/%s.%s.%s.model" % (
            curr_dir, args.language, args.model_type, args.tag_type, args.encoding)
        if args.model_type == "lstm":
            if args.tag_type == "pos":
                model_path = "%s/models/%s/lstm/" % (curr_dir, args.language)
            elif args.tag_type == "chunk":
                model_path = "%s/models/%s/lstm/chunk/" % (
                    curr_dir, args.language)
            elif args.tag_type == "ner":
                model_path = "%s/models/%s/lstm/ner/" % (
                    curr_dir, args.language)
    if args.tag_type != "parse":
        if not os.path.exists(model_path):
            os.makedirs(model_path)

    if args.pipeline_type == 'train':
        logger.info('Start Training#')
        logger.info('Tagger model type: %s' % (args.model_type))
        data_path = "%s/data/train/%s/train.%s.%s" % (
            curr_dir, args.language, args.encoding, args.data_format)
        if args.tag_type == "ner":
            data_path = data_path+".ner"

        data_sents = data_reader.load_data(
            args.data_format, data_path, args.language)

        no_words = sum(len(sent) for sent in data_sents)
        logger.info("No. of words: %d" % (no_words))
        logger.info("No. of sents: %d" % (len(data_sents)))

        X_data = [generate_features.sent2features(
            s, args.tag_type, args.model_type) for s in data_sents]
        y_data = [generate_features.sent2labels(
            s, args.tag_type) for s in data_sents]

        X_train, X_test, y_train, y_test = train_test_split(
            X_data, y_data, test_size=0.10, random_state=42)

        print('Train data size:', len(X_train), len(y_train))
        print('Test data size:', len(X_test), len(y_test))
        print('Lang:', args.language)
        print('Train data: ', data_path)
        print('Model Path: ', model_path)
        if args.model_type == "crf":
            tagger = CRF(model_path)
            tagger.train(X_train, y_train)
            tagger.load_model()
            tagger.test(X_test, y_test)
        elif args.model_type == "lstm":
            x_data, y_data1, y_data2 = load_data_and_labels(data_path)
            if args.tag_type == "pos":
                x_train, x_test, y_train1, y_test1 = train_test_split(
                    x_data, y_data1, test_size=0.10, random_state=42)  # Split the data into train and test
                model = Sequence()  # Intialize BiLSTM model
                # Train the model for 10 echos
                model.fit(x_train, y_train1, epochs=10)
                # Run the model on test data
                print(model.score(x_test, y_test1))
                model.save(model_path+"/weights.h5", model_path +
                           "/params.json", model_path+"/preprocessor.json")
            if args.tag_type == "chunk":
                x_train, x_test, y_train2, y_test2 = train_test_split(
                    x_data, y_data2, test_size=0.10, random_state=42)  # Split the data into train and test
                model = Sequence()  # Intialize BiLSTM model
                # Train the model for 10 echos
                model.fit(x_train, y_train2, epochs=10)
                # Run the model on test data
                print(model.score(x_test, y_test2))
                model.save(model_path+"/weights.h5", model_path +
                           "/params.json", model_path+"/preprocessor.json")
            if args.tag_type == "ner":
                x_train, x_test, y_train1, y_test1 = train_test_split(
                    x_data, y_data1, test_size=0.10, random_state=42)  # Split the data into train and test
                model = Sequence()  # Intialize BiLSTM model
                # Train the model for 10 echos
                model.fit(x_train, y_train1, epochs=10)
                # Run the model on test data
                print(model.score(x_test, y_test1))
                model.save(model_path+"/weights.h5", model_path +
                           "/params.json", model_path+"/preprocessor.json")

    if args.pipeline_type == "test":
        if args.model_type == "crf":
            test_data_path = "%s/%s" % (curr_dir, args.test_data)

            test_sents = data_reader.load_data(
                args.data_format, test_data_path, args.language, tokenize_text=False)
            X_test = [generate_features.sent2features(
                s, args.tag_type, args.model_type) for s in test_sents]
            y_test = [generate_features.sent2labels(
                s, args.tag_type) for s in test_sents]
            tagger = CRF(model_path)
            tagger.load_model()
            tagger.test(X_test, y_test)

    if args.pipeline_type == "predict":

        test_data_path = "%s" % (args.test_data)
        test_sents = data_reader.load_data(
            args.data_format, test_data_path, args.language, tokenize_text=True, split_sent=args.sent_split)
        if args.tag_type == "parse":
            # Pos tagging
            X_test = [generate_features.sent2features(
                s, "pos", args.model_type) for s in test_sents]

            tag_model_path = "%s/models/%s/%s.%s.%s.model" % (
                curr_dir, args.language, args.model_type, "pos", args.encoding)
            chunk_model_path = "%s/models/%s/%s.%s.%s.model" % (
                curr_dir, args.language, args.model_type, "chunk", args.encoding)

            if args.model_type == "crf":
                tagger = CRF(tag_model_path)
                tagger.load_model()
                y_pos = tagger.predict(X_test)

                test_sents_pos = generate_features.append_tags(
                    test_sents, "pos", y_pos)
                X_test = [generate_features.sent2features(
                    s, "chunk", args.model_type) for s in test_sents_pos]

                chunker = CRF(chunk_model_path)
                chunker.load_model()
                y_chunk = chunker.predict(X_test)

                test_fname = path.basename(test_data_path)
                output_file = "%s/%s.parse" % (output_dir, test_fname)
                data_writer.write_anno_to_file(
                    output_file, test_sents_pos, y_chunk, "chunk")
                logger.info("Output in: %s" % output_file)
                data_writer.write_to_screen(output_file)
        else:
            X_test = [generate_features.sent2features(
                s, args.tag_type, args.model_type) for s in test_sents]

            if args.model_type == "crf":
                tagger = CRF(model_path)
                tagger.load_model()
                y_pred = tagger.predict(X_test)
                data_writer.write_anno_to_file(
                    args.output_path, test_sents, y_pred, args.tag_type)
                data_writer.write_to_screen(args.output_path)
                logger.info("Output in: %s" % args.output_path)

            if args.model_type == "lstm":
                model = Sequence().load(model_path+"/weights.h5", model_path +
                                        "/params.json", model_path+"/preprocessor.json")
                f = open(args.test_data, "r")
                sent = f.read()
                tok = Tokenizer(lang=args.language, split_sen=True)
                tokenized_sents = tok.tokenize(sent)
                for tokens in tokenized_sents:
                    for token in tokens:
                        sent = sent + " " + token
                    sent = sent.strip()
                    print(model.analyze(sent))


if __name__ == '__main__':
    pipeline()



from hindiwsd import wsd, lesks


def wordsense(sentence):

    hinglish, hindi = wsd.preprocess_transliterate(sentence)
    print("Hindi Sentence:", hindi)
    pos = wsd.POS_tagger(hindi)

    for i in pos:
        lesks.lesk(i[0], i[1], hindi)

#wordsense("sentence here")





def crf_pos_features(sent, i):
    word = sent[i][0]
    postag = sent[i][1]
    features = [
        'bias',
        'word=' + word,                         #current word
        'word[-4:]=' + word[-4:],               #last 4 characters 
        'word[-3:]=' + word[-3:],               #last 3 characters
        'word[-2:]=' + word[-2:],               #last two characters
        'word.isdigit=%s' % word.isdigit(),     #is a digit
    ]
    if len(word) > 3:
        features.extend([
            'word.short=False'
        ])
    if len(word) < 3:
        features.extend([
            'word.short=True'
        ])
    if i > 0:
        word1 = sent[i-1][0]                                      
        features.extend([
            '-1:word=' + word1,                 #previous word
        ])
    else:
        features.append('BOS')
        
    if i < len(sent)-1:
        word1 = sent[i+1][0]
        postag1 = sent[i+1][1]
        features.extend([
            '+1:word=' + word1,                 #next word
        ])
    else:
        features.append('EOS')
                
    return features


#!/usr/bin/env python
# coding: utf8
from __future__ import unicode_literals, print_function

import plac
import random
from pathlib import Path
import spacy
from spacy.util import minibatch, compounding
import re
import os
import hindiwsd.spacy_data_reader as spacy_data_reader
import argparse
import sys, os.path as path



def get_args():
    ''' This function parses and return arguments passed in'''
    parser = argparse.ArgumentParser(description='Scorer pipeline')

    parser.add_argument("-l", "--language", dest="language", type=str, metavar='<str>', required=True,
                         help="Language of the dataset: tel (telugu), hin (hindi), tam (tamil), kan (kannada), pun (pubjabi)")
    parser.add_argument("-t", "--tag_type", dest="tag_type", type=str, metavar='<str>', required=True,
                         help="Tag type: pos, chunk, parse")
    parser.add_argument("-e", "--encoding", dest="encoding", type=str, metavar='<str>', required=False,
                        help="Encoding of the data (utf, wx)",
                        default="utf")
    parser.add_argument("-v", "--version", dest="version", type=str, metavar='<str>', required=False,
                        help="Version",
                        default="0.0.0")
    return parser.parse_args()

def pipeline():
    args = get_args()
    curr_dir = path.dirname(path.abspath(__file__))
    lang = args.language
    #output_dir = "%s/spacypackages/%s/%s" % (curr_dir, args.language,  args.tag_type)    
    output_dir = "%s/spacypackages/%s_model-%s/%s_model/%s_model-%s" % (curr_dir, args.language,  args.version, args.language, args.language,  args.version)   
    test_text = input("Enter Sentence:")
    #test_text = "నా నా కధ అందరి అందరి ఆడపిల్లల కధే ."
    nlp2 = spacy.load(output_dir)
    doc = nlp2(test_text)
    print('Tags', [(t.text, t.tag_, t.pos_) for t in doc])

if __name__ == '__main__':
    pipeline()



"""Training-related module.
"""
from hindiwsd.callbacks import F1score
from hindiwsd.utils import NERSequence


class Trainer(object):
    """A trainer that train the model.

    Attributes:
        _model: Model.
        _preprocessor: Transformer. Preprocessing data for feature extraction.
    """

    def __init__(self, model, preprocessor=None):
        self._model = model
        self._preprocessor = preprocessor

    def train(self, x_train, y_train, x_valid=None, y_valid=None,
              epochs=1, batch_size=32, verbose=1, callbacks=None, shuffle=True):
        """Trains the model for a fixed number of epochs (iterations on a dataset).

        Args:
            x_train: list of training data.
            y_train: list of training target (label) data.
            x_valid: list of validation data.
            y_valid: list of validation target (label) data.
            batch_size: Integer.
                Number of samples per gradient update.
                If unspecified, `batch_size` will default to 32.
            epochs: Integer. Number of epochs to train the model.
            verbose: Integer. 0, 1, or 2. Verbosity mode.
                0 = silent, 1 = progress bar, 2 = one line per epoch.
            callbacks: List of `keras.callbacks.Callback` instances.
                List of callbacks to apply during training.
            shuffle: Boolean (whether to shuffle the training data
                before each epoch). `shuffle` will default to True.
        """

        train_seq = NERSequence(x_train, y_train, batch_size, self._preprocessor.transform)

        if x_valid and y_valid:
            valid_seq = NERSequence(x_valid, y_valid, batch_size, self._preprocessor.transform)
            f1 = F1score(valid_seq, preprocessor=self._preprocessor)
            callbacks = [f1] + callbacks if callbacks else [f1]

        self._model.fit_generator(generator=train_seq,
                                  epochs=epochs,
                                  callbacks=callbacks,
                                  verbose=verbose,
                                  shuffle=shuffle)



"""
Model API.
"""
import numpy as np
from seqeval.metrics.sequence_labeling import get_entities


class Tagger(object):
    """A model API that tags input sentence.

    Attributes:
        model: Model.
        preprocessor: Transformer. Preprocessing data for feature extraction.
        tokenizer: Tokenize input sentence. Default tokenizer is `str.split`.
    """

    def __init__(self, model, preprocessor, tokenizer=str.split):
        self.model = model
        self.preprocessor = preprocessor
        self.tokenizer = tokenizer

    def predict_proba(self, text):
        """Probability estimates.

        The returned estimates for all classes are ordered by the
        label of classes.

        Args:
            text : string, the input text.

        Returns:
            y : array-like, shape = [num_words, num_classes]
            Returns the probability of the word for each class in the model,
        """
        assert isinstance(text, str)

        words = self.tokenizer(text)
        X = self.preprocessor.transform([words])
        y = self.model.predict(X)
        y = y[0]  # reduce batch dimension.

        return y

    def _get_prob(self, pred):
        prob = np.max(pred, -1)

        return prob

    def _get_tags(self, pred):
        tags = self.preprocessor.inverse_transform([pred])
        tags = tags[0]  # reduce batch dimension

        return tags

    def _build_response(self, sent, tags, prob):
        words = self.tokenizer(sent)
        res = {
            'words': words,
            'entities': [

            ]
        }
        chunks = get_entities(tags)

        for chunk_type, chunk_start, chunk_end in chunks:
            chunk_end += 1
            entity = {
                'text': ' '.join(words[chunk_start: chunk_end]),
                'type': chunk_type,
                'score': float(np.average(prob[chunk_start: chunk_end])),
                'beginOffset': chunk_start,
                'endOffset': chunk_end
            }
            res['entities'].append(entity)

        return res

    def _build_response1(self, sent, tags, prob):
        words = self.tokenizer(sent)
        res = "" 
        chunks = get_entities(tags)
        for index, obj in enumerate(words):
            res = res + obj  +"\t"+tags[index] +"\n"
            if "." in obj:
                  res = res+"\n"
            if "।" in obj:
                  res = res+"\n"

        return res
    def analyze(self, text):

        pred = self.predict_proba(text)
        tags = self._get_tags(pred)
        prob = self._get_prob(pred)
        res = self._build_response1(text, tags, prob)
        return res

    def predict(self, text):
        pred = self.predict_proba(text)
        tags = self._get_tags(pred)

        return tags



"""
Model definition.
"""
import json

from tensorflow.keras.layers import Dense, LSTM, Bidirectional, Embedding, Input, Dropout, TimeDistributed
from tensorflow.keras.layers import Concatenate
from tensorflow.keras.models import Model, model_from_json

from hindiwsd.layers import CRF


def save_model(model, weights_file, params_file):
    with open(params_file, 'w') as f:
        params = model.to_json()
        json.dump(json.loads(params), f, sort_keys=True, indent=4)
        model.save_weights(weights_file)


def load_model(weights_file, params_file):
    with open(params_file) as f:
        model = model_from_json(f.read(), custom_objects={'CRF': CRF})
        model.load_weights(weights_file)

    return model


class BiLSTMCRF(object):
    """A Keras implementation of BiLSTM-CRF for sequence labeling.

    References
    --
    Guillaume Lample, Miguel Ballesteros, Sandeep Subramanian, Kazuya Kawakami, Chris Dyer.
    "Neural Architectures for Named Entity Recognition". Proceedings of NAACL 2016.
    https://arxiv.org/abs/1603.01360
    """

    def __init__(self,
                 num_labels,
                 word_vocab_size,
                 char_vocab_size=None,
                 word_embedding_dim=100,
                 char_embedding_dim=25,
                 word_lstm_size=100,
                 char_lstm_size=25,
                 fc_dim=100,
                 dropout=0.5,
                 embeddings=None,
                 use_char=True,
                 use_crf=True):
        """Build a Bi-LSTM CRF model.

        Args:
            word_vocab_size (int): word vocabulary size.
            char_vocab_size (int): character vocabulary size.
            num_labels (int): number of entity labels.
            word_embedding_dim (int): word embedding dimensions.
            char_embedding_dim (int): character embedding dimensions.
            word_lstm_size (int): character LSTM feature extractor output dimensions.
            char_lstm_size (int): word tagger LSTM output dimensions.
            fc_dim (int): output fully-connected layer size.
            dropout (float): dropout rate.
            embeddings (numpy array): word embedding matrix.
            use_char (boolean): add char feature.
            use_crf (boolean): use crf as last layer.
        """
        super(BiLSTMCRF).__init__()
        self._char_embedding_dim = char_embedding_dim
        self._word_embedding_dim = word_embedding_dim
        self._char_lstm_size = char_lstm_size
        self._word_lstm_size = word_lstm_size
        self._char_vocab_size = char_vocab_size
        self._word_vocab_size = word_vocab_size
        self._fc_dim = fc_dim
        self._dropout = dropout
        self._use_char = use_char
        self._use_crf = use_crf
        self._embeddings = embeddings
        self._num_labels = num_labels

    def build(self):
        # build word embedding
        word_ids = Input(batch_shape=(None, None), dtype='int32', name='word_input')
        inputs = [word_ids]
        if self._embeddings is None:
            word_embeddings = Embedding(input_dim=self._word_vocab_size,
                                        output_dim=self._word_embedding_dim,
                                        mask_zero=True,
                                        name='word_embedding')(word_ids)
        else:
            word_embeddings = Embedding(input_dim=self._embeddings.shape[0],
                                        output_dim=self._embeddings.shape[1],
                                        mask_zero=True,
                                        weights=[self._embeddings],
                                        name='word_embedding')(word_ids)

        # build character based word embedding
        if self._use_char:
            char_ids = Input(batch_shape=(None, None, None), dtype='int32', name='char_input')
            inputs.append(char_ids)
            char_embeddings = Embedding(input_dim=self._char_vocab_size,
                                        output_dim=self._char_embedding_dim,
                                        mask_zero=True,
                                        name='char_embedding')(char_ids)
            char_embeddings = TimeDistributed(Bidirectional(LSTM(self._char_lstm_size)))(char_embeddings)
            word_embeddings = Concatenate()([word_embeddings, char_embeddings])

        word_embeddings = Dropout(self._dropout)(word_embeddings)
        z = Bidirectional(LSTM(units=self._word_lstm_size, return_sequences=True))(word_embeddings)
        z = Dense(self._fc_dim, activation='tanh')(z)

        if self._use_crf:
            crf = CRF(self._num_labels, sparse_target=False)
            loss = crf.loss_function
            pred = crf(z)
        else:
            loss = 'categorical_crossentropy'
            pred = Dense(self._num_labels, activation='softmax')(z)

        model = Model(inputs=inputs, outputs=pred)

        return model, loss


class ELModel(object):
    """
    A Keras implementation of ELMo BiLSTM-CRF for sequence labeling.
    """

    def __init__(self,
                 num_labels,
                 word_vocab_size,
                 char_vocab_size=None,
                 word_embedding_dim=100,
                 char_embedding_dim=25,
                 word_lstm_size=100,
                 char_lstm_size=25,
                 fc_dim=100,
                 dropout=0.5,
                 embeddings=None):
        """Build a Bi-LSTM CRF model.

        Args:
            word_vocab_size (int): word vocabulary size.
            char_vocab_size (int): character vocabulary size.
            num_labels (int): number of entity labels.
            word_embedding_dim (int): word embedding dimensions.
            char_embedding_dim (int): character embedding dimensions.
            word_lstm_size (int): character LSTM feature extractor output dimensions.
            char_lstm_size (int): word tagger LSTM output dimensions.
            fc_dim (int): output fully-connected layer size.
            dropout (float): dropout rate.
            embeddings (numpy array): word embedding matrix.
        """
        self._char_embedding_dim = char_embedding_dim
        self._word_embedding_dim = word_embedding_dim
        self._char_lstm_size = char_lstm_size
        self._word_lstm_size = word_lstm_size
        self._char_vocab_size = char_vocab_size
        self._word_vocab_size = word_vocab_size
        self._fc_dim = fc_dim
        self._dropout = dropout
        self._embeddings = embeddings
        self._num_labels = num_labels

    def build(self):
        # build word embedding
        word_ids = Input(batch_shape=(None, None), dtype='int32', name='word_input')
        if self._embeddings is None:
            word_embeddings = Embedding(input_dim=self._word_vocab_size,
                                        output_dim=self._word_embedding_dim,
                                        mask_zero=True,
                                        name='word_embedding')(word_ids)
        else:
            word_embeddings = Embedding(input_dim=self._embeddings.shape[0],
                                        output_dim=self._embeddings.shape[1],
                                        mask_zero=True,
                                        weights=[self._embeddings],
                                        name='word_embedding')(word_ids)

        # build character based word embedding
        char_ids = Input(batch_shape=(None, None, None), dtype='int32', name='char_input')
        char_embeddings = Embedding(input_dim=self._char_vocab_size,
                                    output_dim=self._char_embedding_dim,
                                    mask_zero=True,
                                    name='char_embedding')(char_ids)
        char_embeddings = TimeDistributed(Bidirectional(LSTM(self._char_lstm_size)))(char_embeddings)

        elmo_embeddings = Input(shape=(None, 1024), dtype='float32')

        word_embeddings = Concatenate()([word_embeddings, char_embeddings, elmo_embeddings])

        word_embeddings = Dropout(self._dropout)(word_embeddings)
        z = Bidirectional(LSTM(units=self._word_lstm_size, return_sequences=True))(word_embeddings)
        z = Dense(self._fc_dim, activation='tanh')(z)

        crf = CRF(self._num_labels, sparse_target=False)
        loss = crf.loss_function
        pred = crf(z)

        model = Model(inputs=[word_ids, char_ids, elmo_embeddings], outputs=pred)

        return model, loss




from statistics import mean
from indic_transliteration import sanscript
from indic_transliteration.sanscript import SchemeMap, SCHEMES, transliterate

from spello.model import SpellCorrectionModel
import pyiwn
import sys

from nltk.tag import tnt
from nltk.corpus import indian
import nltk

from hindiwsd import wsd
import pandas as pd

def lesk(word, postag, sentence):

    # sentence - raw hinglish sentence to be transliterated to hindi and wsd carried out
    # returns each word and it's respective meaning

    # Transliterate
    # corrected_hindi = wsd.preprocess_transliterate(sentence)

    # POS tagging for words in the transliterated sentence
    # tags = wsd.POS_tagger(sentence)

    iwn = pyiwn.IndoWordNet()

    # Function for enhancing lesk
    
    def extra_overlap(ambword : str, input_sent : str):

        df = pd.read_csv('./Hindi_WSD_Dataset - Sheet1.tsv', sep = '\t', encoding='utf8')
        #df.head()
        #print(ambword)

        word1 = list(df['Word'])
        sentence = list(df['Sentence'])
        synset = list(df['synset number'])
        words = [] #the words
        sent = [] #their particular sentences
        syn = []
  
        # print(word1)
        for i in range(len(word1)):
            #print(word1[i])
            if str(ambword) == str(word1[i]):
                sent.append(sentence[i])
                syn.append(synset[i])

            # elif str(ambword) != str(word1[i]):

        
        #print("Sentence:", sent)
        lis = list(input_sent.split())
        max_intersection = 0
        synset_final = -1
        meaning = ""

        for i in range(len(syn)):
            input_set = set()
            
            for x in lis:
                #print(x)
                input_set.add(x)
            
            sent_lis = sent[i].split()
            
            # print("Sent_Lis:", sent_lis)
            # print("Input_Set:", input_set)
            
            overlap = input_set.intersection(sent_lis)
            
            #print("Overlap:", overlap)
            
            # synset_final = 0
            if len(overlap) > max_intersection:
                max_intersection = len(overlap)
                synset_final = syn[i]
                if synset_final.startswith("N1"):
                    meaning = synset_final[3:]
            else:
                continue 

        
        return max_intersection, synset_final, meaning

    
    def scanLine(line, wordset):
        strings = line.split()

        for s in strings:
            size = len(s)
            if size <= 2:
                continue
            else:
                wordset[s] = 1


    def computeOverlap(sign, context):
        overlap = 0
        for word, val in sign.items():
            v = context.get(word, 0)
            if v != 0:
                overlap += 1
        return overlap

    #################### Lesk's Algorithm ####################

    context = {}
    scanLine(sentence, context)

    if postag == "NOUN":
        senses = iwn.synsets(word, pos=pyiwn.PosTag.NOUN)
    elif postag == "VERB":
        senses = iwn.synsets(word, pos=pyiwn.PosTag.VERB)
    elif postag == "ADVERB":
        senses = iwn.synsets(word, pos=pyiwn.PosTag.ADVERB)
    elif postag == "ADJECTIVE":
        senses = iwn.synsets(word, pos=pyiwn.PosTag.ADJECTIVE)

    if len(senses) >= 0:
        max_overlap = -1
        result = ''

        for s in senses:
            info = s.gloss()
            for e in s.examples():
                info = info + ' ' + e

            signature = {}

            scanLine(info, signature)
            overlap = computeOverlap(signature, context)
            if overlap > max_overlap:
                max_overlap = overlap
                result = s.gloss()

        intersection, syn_final, meaning = extra_overlap(word, sentence)

        #print("MAX OVERLAP:", max_overlap, "INTERSECTION:", intersection, "SYNSET:", syn_final)
        
        if str(syn_final).startswith("N1"):
            result = meaning
        
        else:
            if intersection>=max_overlap and intersection>0:
                result = iwn.synsets(word)[int(syn_final)].gloss()
                #result = senses[int(syn_final)].gloss()

            if result=="":
                result = "NOT available in wordnet"

        print("Meaning of", word, " is:", result)

    else:
        print(word, "is not in wordnet")







import re, os
import codecs
import logging
from polyglot_tokenizer import Tokenizer

logger = logging.getLogger(__name__)

def load_data(text_type, filename, lang, tokenize_text=False, split_sent=True):
    data_tuple = []
    with codecs.open(filename, 'r', encoding='utf-8') as fp:
        logger.info('Loading text_type: %s format' % (text_type))
        if text_type == "ssf":
            start_c = -1
            for line in fp:
                line = line.strip()
                ds = line.split()
                #print("Line", line)
                #print("DS", ds)
                if line == "":
                    continue
                elif line[0:2] == "<S":
                    sent = []
                elif line[0:3] == "</S":
                    data_tuple.append(sent)
                elif line[0] == "<":
                    continue
                elif ds[0] == "0" or ds[0] == "))":
                    continue
                elif ds[1] == "((":
                    start_c, chunk_tag = 1, ds[2]
                    #print "hello-chunk tag",chunk_tag
                if len(ds) > 2:
                    if ds[2]:
                        #print "--",line,"--"
                        word, tag = ds[1], ds[2]
                        if start_c == -1:
                            sent.append((word, tag, ""))
                        if start_c == 1:
                            sent.append((word, tag, "B-%s" % (chunk_tag)))
                            start_c = 0
                        if start_c == 0:
                            sent.append((word, tag, "I-%s" % (chunk_tag)))
        elif text_type == "conll":
            sent = []
            for line in fp:
                line = line.strip()
                ds = line.split()
                if line != "":
                    print(line)
                    if len(ds) == 2:
                        word, tag, chunk = ds[1], "",""
                    if len(ds) == 3:
                        word, tag, chunk = ds[1], ds[2], ""
                    if len(ds) == 4:
                        word, tag, chunk = ds[1], ds[2], ds[3]
                    sent.append([word, tag, chunk])
                else:
                    data_tuple.append(sent)
                    sent = []
        elif text_type == "txt":
            if split_sent == True:
                text = fp.read()
                tok = Tokenizer(lang=lang, split_sen=split_sent)
                tokenized_sents = tok.tokenize(text)
                sent = []
                for tokens in tokenized_sents:
                    for token in tokens:
                        sent.append([token, "", ""])
                    data_tuple.append(sent)
            else:
                for line in fp:
                    sent = []
                    if tokenize_text:
                        tok = Tokenizer(lang=lang, split_sen=False)
                        tokenized_sents = tok.tokenize(line)
                    for tokens in tokenized_sents:
                        for token in tokens:
                            sent.append([token, "", ""])
                    data_tuple.append(sent)
        else:
            print("Check - text_type", text_type)

    return data_tuple



import sys,os.path as path
sys.path.append(path.dirname(path.dirname(path.dirname(path.abspath(__file__)))))

"""
Convert SSF To CONLL format

python convert_ssf_conll.py -i ../../data/test/tel/test.utf.ssf.pos -o ../../data/test/tel/test.utf.conll.pos

SSF format:
-----------

<Sentence id=1>
1	((	chunk11
1.1 word11	tag11
1.2	word12 	tag12		
	))
2	((	chunk12
2.1 word13 tag13
	))	
</Sentence>
<Sentence id=2>
1	((	chunk11
1.1 word11	tag11
1.2	word12 	tag12		
1.3 word13 tag13
	))	
</Sentence>
1 word11 tag11 B-chunk11
2 
3 word13 tag13 B-chunk13

CONLL format: (BI format)
-------------------------
1 word11 tag11 B-chunk11
2 word12 tag12 I-chunk11
3 word13 tag13 B-chunk12

1 word21 tag21 B-chunk21
2 word22 tag22 I-chunk22
3 word23 tag23 I-chunk23

"""

import re, os
import codecs
import logging
from tagger.src import data_reader
from tagger.utils.writer import write_to_file
import argparse
logger = logging.getLogger(__name__)

def convert_format(sents, filename):
	with codecs.open(filename, 'w', encoding='utf8', errors='ignore') as fp:
		for sent in sents:
			for index, item in enumerate(sent):
				tag = item[1].replace("?","").replace(":","")
				chunk = item[2].replace("?","").replace(":","")
				word = item[0].replace("?","").replace(":","")
				if tag:
					if word:
						fp.write(str(index+1) + "\t" + word +"\t" + tag+"\t"+chunk + "\n")
			fp.write("\n")

def get_args():
    ''' This function parses and return arguments passed in'''
    parser = argparse.ArgumentParser(description='Format Converter from SSF to CONLL')
    parser.add_argument("-i", "--input_path", dest="file_path", type=str, metavar='<str>', required=True,
                        help="File to be converted file path ex: data/test/telugu/test.wx.conll")
    parser.add_argument("-o", "--output_path", dest="output_path", type=str, metavar='<str>', required=True,
                         help="The path to the output file ex: outputs/test.utf.conll")
    return parser.parse_args()


if __name__ == "__main__":

	curr_dir = path.dirname(path.abspath(__file__))
	args = get_args()

	filename = "%s/%s" % (curr_dir, args.file_path)
	output_file = "%s/%s" % (curr_dir, args.output_path)
	sents = data_reader.load_data("ssf", filename, "")
	convert_format(sents, output_file)













def crf_chunk_features(sent, i):
    word = sent[i][0]
    postag = sent[i][1]
    features = [
        'bias',
        'word=' + word,                             #current word
        'word[-4:]=' + word[-4:],                   #last 4 characters 
        'word[-3:]=' + word[-3:],                   #last 3 characters
        'word[-2:]=' + word[-2:],                   #last 2 characters
        'word.isdigit=%s' % word.isdigit(),         #is a digit?
        'postag=' + postag,                         #current POS tag
        'postag[:2]=' + postag[:2],                 #first two characters of POS tag
    ]
    if len(word) > 3:
        features.extend([
            'word.short=False'
        ])
    if len(word) <= 3:
        features.extend([
            'word.short=True'
        ])
    if i > 0:
        word1 = sent[i-1][0]
        postag1 = sent[i-1][1]
        features.extend([
            '-1:word=' + word1,                     #previous word
            '-1:postag=' + postag1,                 #previous POS tag
            '-1:postag[:2]=' + postag1[:2],         #first two characters of previous POS tag
        ])
    else:
        features.append('BOS')
        
    if i < len(sent)-1:
        word1 = sent[i+1][0]
        postag1 = sent[i+1][1]
        features.extend([
            '+1:word=' + word1,                     #next word
            '+1:postag=' + postag1,                 #next POS tag
            '+1:postag[:2]=' + postag1[:2],         #first two characters of next POS tag 
        ])
    else:
        features.append('EOS')
                
    return features


from polyglot_tokenizer import Tokenizer
from hindiwsd.CRF import CRF 
from sklearn.model_selection import train_test_split
from time import time
import numpy as np
import pickle
import logging
import argparse
import hindiwsd.writer as data_writer
from hindiwsd import generate_features, data_reader 
#import lstmcrf
from hindiwsd.utils import load_data_and_labels
from hindiwsd.wrapper import Sequence
import sys
import os.path as path
import os
sys.path.append(path.dirname(path.abspath(__file__)))


def pos_pipeline(sentence):

    f = open('./test.txt', 'w', encoding='utf8')
    f.write(sentence)
    f.close()

    curr_dir = path.dirname(path.abspath(__file__))
    args = {'pipeline_type': 'predict', 'language': 'hi', 'tag_type': 'pos',
            'model_type': 'crf', 'encoding': 'utf', 'data_format': 'txt', 'sent_split': True, 'test_data': './test.txt', 'output_path': './out.txt'}

    output_dir = path.join(path.dirname(path.abspath(__file__)), "outputs")
    # if not os.path.exists(output_dir):
    #     os.makedirs(output_dir)

    # data_writer.set_logger(args.model_type, output_dir)

    if True:
        model_path = "%s/indic_tagger/models/%s/%s.%s.%s.model" % (
            curr_dir, args['language'], args['model_type'], args['tag_type'], args['encoding'])
        if args['model_type'] == "lstm":
            if args['tag_type'] == "pos":
                model_path = "%s/models/%s/lstm/" % (
                    curr_dir, args['language'])
            elif args['tag_type'] == "chunk":
                model_path = "%s/models/%s/lstm/chunk/" % (
                    curr_dir, args['language'])
            elif args['tag_type'] == "ner":
                model_path = "%s/models/%s/lstm/ner/" % (
                    curr_dir, args['language'])
    # if args['tag_type'] != "parse":
    #     if not os.path.exists(model_path):
    #         os.makedirs(model_path)

    if args['pipeline_type'] == "predict":

        test_data_path = "%s" % (args['test_data'])
        test_sents = data_reader.load_data(
            args['data_format'], test_data_path, args['language'], tokenize_text=True, split_sent=args['sent_split'])
        if args['tag_type'] == "parse":
            # Pos tagging
            X_test = [generate_features.sent2features(
                s, "pos", args['model_type']) for s in test_sents]

            tag_model_path = "%s/indic_tagger/models/%s/%s.%s.%s.model" % (
                curr_dir, args['language'], args['model_type'], "pos", args['encoding'])
            chunk_model_path = "%s/indic_tagger/models/%s/%s.%s.%s.model" % (
                curr_dir, args['language'], args['model_type'], "chunk", args['encoding'])

            if args['model_type'] == "crf":
                tagger = CRF('crf.pos.utf.model')
                tagger.load_model()
                y_pos = tagger.predict(X_test)

                test_sents_pos = generate_features.append_tags(
                    test_sents, "pos", y_pos)
                X_test = [generate_features.sent2features(
                    s, "chunk", args['model_type']) for s in test_sents_pos]

                chunker = CRF(chunk_model_path)
                chunker.load_model()
                y_chunk = chunker.predict(X_test)

                test_fname = path.basename(test_data_path)
                output_file = "%s/%s.parse" % (output_dir, test_fname)
                data_writer.write_anno_to_file(
                    output_file, test_sents_pos, y_chunk, "chunk")
                # logger.info("Output in: %s" % output_file)
                data_writer.write_to_screen(output_file)
        else:
            X_test = [generate_features.sent2features(
                s, args['tag_type'], args['model_type']) for s in test_sents]

            if args['model_type'] == "crf":
                tagger = CRF('crf.pos.utf.model')
                tagger.load_model()
                y_pred = tagger.predict(X_test)
                # print("Y_PRED:", y_pred)
                data_writer.write_anno_to_file(
                    args['output_path'], test_sents, y_pred, args['tag_type'])
                tagged = data_writer.write_to_screen(args['output_path'])
                # logger.info("Output in: %s" % args.output_path)

    return tagged.split()


# tagged = pos_pipeline("कल मैं बाहर जाऊंगा")
# print(tagged)



"""
Custom callbacks.
"""
import numpy as np
from keras.callbacks import Callback
from seqeval.metrics import f1_score, classification_report


class F1score(Callback):

    def __init__(self, seq, preprocessor=None):
        super(F1score, self).__init__()
        self.seq = seq
        self.p = preprocessor

    def get_lengths(self, y_true):
        lengths = []
        for y in np.argmax(y_true, -1):
            try:
                i = list(y).index(0)
            except ValueError:
                i = len(y)
            lengths.append(i)

        return lengths

    def on_epoch_end(self, epoch, logs={}):
        label_true = []
        label_pred = []
        for i in range(len(self.seq)):
            x_true, y_true = self.seq[i]
            lengths = self.get_lengths(y_true)
            y_pred = self.model.predict_on_batch(x_true)

            y_true = self.p.inverse_transform(y_true, lengths)
            y_pred = self.p.inverse_transform(y_pred, lengths)

            label_true.extend(y_true)
            label_pred.extend(y_pred)

        score = f1_score(label_true, label_pred)
        print(' - f1: {:04.2f}'.format(score * 100))
        print(classification_report(label_true, label_pred))
        logs['f1'] = score



"""
Utility functions.
"""
import math
import os
from collections import Counter

import numpy as np
from tensorflow.keras.utils import Sequence, get_file


def download(url):
    """Download a trained weights, config and preprocessor.

    Args:
        url (str): target url.
    """
    filepath = get_file(fname='tmp.zip', origin=url, extract=True)
    base_dir = os.path.dirname(filepath)
    weights_file = os.path.join(base_dir, 'weights.h5')
    params_file = os.path.join(base_dir, 'params.json')
    preprocessor_file = os.path.join(base_dir, 'preprocessor.pickle')

    return weights_file, params_file, preprocessor_file


def load_data_and_labels(filename, encoding='utf-8'):
    sents, labels, clabels = [], [], []
    words, tags, chunks = [], [], []
    with open(filename, encoding=encoding) as f:
        for line in f:
            line = line.rstrip()
            if line:
                temp, word, tag, chunk = line.split('\t')
                words.append(word)
                tags.append(tag)
                chunks.append(chunk)
            else:
                if words:
                    sents.append(words)
                    labels.append(tags)
                    clabels.append(chunks)
                words, tags, chunks = [], [], []

    return sents, labels, clabels


class NERSequence(Sequence):

    def __init__(self, x, y, batch_size=1, preprocess=None):
        self.x = x
        self.y = y
        self.batch_size = batch_size
        self.preprocess = preprocess

    def __getitem__(self, idx):
        batch_x = self.x[idx * self.batch_size: (idx + 1) * self.batch_size]
        batch_y = self.y[idx * self.batch_size: (idx + 1) * self.batch_size]

        return self.preprocess(batch_x, batch_y)

    def __len__(self):
        return math.ceil(len(self.x) / self.batch_size)


class Vocabulary(object):
    """A vocabulary that maps tokens to ints (storing a vocabulary).

    Attributes:
        _token_count: A collections.Counter object holding the frequencies of tokens
            in the data used to build the Vocabulary.
        _token2id: A collections.defaultdict instance mapping token strings to
            numerical identifiers.
        _id2token: A list of token strings indexed by their numerical identifiers.
    """

    def __init__(self, max_size=None, lower=True, unk_token=True, specials=('<pad>',)):
        """Create a Vocabulary object.

        Args:
            max_size: The maximum size of the vocabulary, or None for no
                maximum. Default: None.
            lower: boolean. Whether to convert the texts to lowercase.
            unk_token: boolean. Whether to add unknown token.
            specials: The list of special tokens (e.g., padding or eos) that
                will be prepended to the vocabulary. Default: ('<pad>',)
        """
        self._max_size = max_size
        self._lower = lower
        self._unk = unk_token
        self._token2id = {token: i for i, token in enumerate(specials)}
        self._id2token = list(specials)
        self._token_count = Counter()

    def __len__(self):
        return len(self._token2id)

    def add_token(self, token):
        """Add token to vocabulary.

        Args:
            token (str): token to add.
        """
        token = self.process_token(token)
        self._token_count.update([token])

    def add_documents(self, docs):
        """Update dictionary from a collection of documents. Each document is a list
        of tokens.

        Args:
            docs (list): documents to add.
        """
        for sent in docs:
            sent = map(self.process_token, sent)
            self._token_count.update(sent)

    def doc2id(self, doc):
        """Get the list of token_id given doc.

        Args:
            doc (list): document.

        Returns:
            list: int id of doc.
        """
        doc = map(self.process_token, doc)
        return [self.token_to_id(token) for token in doc]

    def id2doc(self, ids):
        """Get the token list.

        Args:
            ids (list): token ids.

        Returns:
            list: token list.
        """
        return [self.id_to_token(idx) for idx in ids]

    def build(self):
        """
        Build vocabulary.
        """
        token_freq = self._token_count.most_common(self._max_size)
        idx = len(self.vocab)
        for token, _ in token_freq:
            self._token2id[token] = idx
            self._id2token.append(token)
            idx += 1
        if self._unk:
            unk = '<unk>'
            self._token2id[unk] = idx
            self._id2token.append(unk)

    def process_token(self, token):
        """Process token before following methods:
        * add_token
        * add_documents
        * doc2id
        * token_to_id

        Args:
            token (str): token to process.

        Returns:
            str: processed token string.
        """
        if self._lower:
            token = token.lower()

        return token

    def token_to_id(self, token):
        """Get the token_id of given token.

        Args:
            token (str): token from vocabulary.

        Returns:
            int: int id of token.
        """
        token = self.process_token(token)
        return self._token2id.get(token, len(self._token2id) - 1)

    def id_to_token(self, idx):
        """token-id to token (string).

        Args:
            idx (int): token id.

        Returns:
            str: string of given token id.
        """
        return self._id2token[idx]

    @property
    def vocab(self):
        """Return the vocabulary.

        Returns:
            dict: get the dict object of the vocabulary.
        """
        return self._token2id

    @property
    def reverse_vocab(self):
        """Return the vocabulary as a reversed dict object.

        Returns:
            dict: reversed vocabulary object.
        """
        return self._id2token


def filter_embeddings(embeddings, vocab, dim):
    """Loads word vectors in numpy array.

    Args:
        embeddings (dict): a dictionary of numpy array.
        vocab (dict): word_index lookup table.

    Returns:
        numpy array: an array of word embeddings.
    """
    if not isinstance(embeddings, dict):
        return
    _embeddings = np.zeros([len(vocab), dim])
    for word in vocab:
        if word in embeddings:
            word_idx = vocab[word]
            _embeddings[word_idx] = embeddings[word]

    return _embeddings


def load_glove(file):
    """Loads GloVe vectors in numpy array.

    Args:
        file (str): a path to a glove file.

    Return:
        dict: a dict of numpy arrays.
    """
    model = {}
    with open(file) as f:
        for line in f:
            line = line.split(' ')
            word = line[0]
            vector = np.array([float(val) for val in line[1:]])
            model[word] = vector

    return model



from indic_transliteration import sanscript
from indic_transliteration.sanscript import SchemeMap, SCHEMES, transliterate

from spello.model import SpellCorrectionModel
import pyiwn
import sys

from nltk.tag import tnt
from nltk.corpus import indian
import nltk
from hindiwsd.hipos import pos_pipeline

# Loading spell correction model
sp = SpellCorrectionModel(language='hi')
sp.load('./hi.pkl')

# Stopwords
# words = []
# with open('/content/drive/MyDrive/hinglish_stopwords.txt', 'r') as f:
#   words = f.readlines()

# stopwords = []
# for word in words:
#   stopwords.append(word[:-1])


def preprocess_transliterate(sentence):

    # sentence - hinglish sentence to be processed
    # Returns hinglish and hindi sentence after stopword removal, transliteration and spell check

    arr = sentence.split()

    # Correcting hinglish words to assist hindi correction model
    barr = []
    vowels = 'aeiou'
    consonants = 'bcdfghjklmnpqrstvwxyz'
    # sarr_new = []

    for word in arr:

        # flag = 0

        # if word in sarr:
        #   flag = 1

        for i in range(1, len(word)-2):
            # and word[i+2] not in consonants:
            if len(word) > 3 and word[i-1] in consonants and word[i+1] in consonants and word[i] == 'a':
                word = word[:i] + 'a' + word[i:]
                break

        # if len(word) == 3:
        #     if word[1] == 'a' and word[0] in consonants and word[2] in consonants:
        #       word = word[0] + word[1] + 'a' + word[2]

        if word[len(word) - 1] == 'a' and word[len(word) - 2] != 'a':
            word = word + 'a'
            barr.append(word)
        else:
            barr.append(word)

        # if flag:
        #     sarr_new.append(word)

    # Transliteration
    hinglish = ' '.join(barr)
    hindi = transliterate(hinglish, sanscript.ITRANS, sanscript.DEVANAGARI)

    # hinglish_swr = ' '.join(sarr_new)
    # hindi_swr = transliterate(hinglish_swr, sanscript.ITRANS, sanscript.DEVANAGARI)

    # Spell correction
    tf = sp.spell_correct(hindi)
    corrected_hindi = tf['spell_corrected_text']

    # tf_swr = sp.spell_correct(hindi_swr)
    # corrected_hindi_swr = tf_swr['spell_corrected_text']

    return hinglish, corrected_hindi  # corrected_hindi_swr


def POS_tagger(sentence):

    # sentence - Hindi sentence for which POS is needed to be found
    # returns mapped POS such that they can be used with pyiwn

    tagged = pos_pipeline(sentence)

    # print("Tagged:", tagged)

    words = []
    tags = []

    for i in range(len(tagged)):

        if i % 2 == 0:
            words.append(tagged[i])
        else:
            tags.append(tagged[i])

    pos = []

    for i in range(len(words)):

        if (tags[i].startswith('N')):
            pos.append((words[i], 'NOUN'))
        elif (tags[i].startswith('V')):
            pos.append((words[i], 'VERB'))
        elif (tags[i].startswith('J') or tags[i].startswith('Q')):
            pos.append((words[i], 'ADJECTIVE'))
        elif (tags[i].startswith('R')):
            pos.append((words[i], 'ADVERB'))

    return pos




import pycrfsuite
from hindiwsd import evaluate


class CRF():
    def __init__(self, model_path):
        self.trainer = pycrfsuite.Trainer(verbose=False)
        self.model_path = model_path
        self.trainer.set_params({
            'c1': 1.0,   # coefficient for L1 penalty
            'c2': 1-3,  # coefficient for L2 penalty
            'max_iterations': 50,  # stop earlier

            # include transitions that are possible, but not observed
            'feature.possible_transitions': True
        })

    def train(self, X_train, y_train):
        for xseq, yseq in zip(X_train, y_train):
            self.trainer.append(xseq, yseq)
        self.trainer.train(self.model_path)

    def load_model(self):
        self.tagger = pycrfsuite.Tagger()
        self.tagger.open(self.model_path)

    def test(self, X_test, y_test):
        y_pred = self.predict(X_test)
        print(evaluate.bio_classification_report(y_test, y_pred))

    def predict(self, X_test):
        y_pred = [self.tagger.tag(xseq) for xseq in X_test]
        return y_pred



#!/usr/bin/env python
# coding: utf8
from __future__ import unicode_literals, print_function

import plac
import random
from pathlib import Path
import spacy
from spacy.util import minibatch, compounding
import re
import os
import hindiwsd.spacy_data_reader as spacy_data_reader
import argparse
import sys, os.path as path

TAG_MAP = {
    'CC': {'pos': 'NOUN'},
    'CL': {'pos': 'NOUN'},
    'DEM': {'pos': 'NOUN'},
    'ECH': {'pos': 'NOUN'},
    'INJ': {'pos': 'NOUN'},
    'INTF': {'pos': 'NOUN'},
    'INTJ': {'pos': 'NOUN'},
    'J': {'pos': 'ADJ'},
    'JJ': {'pos': 'ADJ'},
    'NN': {'pos': 'NOUN'},
    'NNP': {'pos': 'NOUN'},
    'NNPP': {'pos': 'NOUN'},
    'NP': {'pos': 'NOUN'},
    'NST': {'pos': 'NOUN'},
    'PRP': {'pos': 'NOUN'},
    'PSP': {'pos': 'NOUN'},
    'QC': {'pos': 'NOUN'},
    'QF': {'pos': 'NOUN'},
    'QO': {'pos': 'NOUN'},
    'RB': {'pos': 'NOUN'},
    'RDP': {'pos': 'NOUN'},
    'RM': {'pos': 'NOUN'},
    'RP': {'pos': 'NOUN'},
    'SY': {'pos': 'NOUN'},
    'SYM': {'pos': 'SYM'},
    'UNK': {'pos': 'NOUN'},
    'UT': {'pos': 'NOUN'},
    'VAUX': {'pos': 'VERB'},
    'VM': {'pos': 'VERB'},
    'WQ': {'pos': 'NOUN'}
}



def get_args():
    ''' This function parses and return arguments passed in'''
    parser = argparse.ArgumentParser(description='Scorer pipeline')

    parser.add_argument("-l", "--language", dest="language", type=str, metavar='<str>', required=True,
                         help="Language of the dataset: tel (telugu), hin (hindi), tam (tamil), kan (kannada), pun (pubjabi)")
    parser.add_argument("-t", "--tag_type", dest="tag_type", type=str, metavar='<str>', required=True,
                         help="Tag type: pos, chunk, parse")
    parser.add_argument("-e", "--encoding", dest="encoding", type=str, metavar='<str>', required=False,
                        help="Encoding of the data (utf, wx)",
                        default="utf")
    parser.add_argument("-i", "--input_file", dest="test_data", type=str, metavar='<str>', required=False,
                        help="Test data path ex: data/test/telugu/test.txt")
    parser.add_argument("-s", "--sent_split", dest="sent_split", type=str, metavar='<str>', required=False,
                        help="Test data path ex: data/test/telugu/test.txt",
                        default=True)
    parser.add_argument("-o", "--output_file", dest="output_path", type=str, metavar='<str>',
                         help="The path to the output file",
                         default=path.join(path.dirname(path.abspath(__file__)), "outputs", "output_file"))
    return parser.parse_args()

def pipeline():
    """Create a new model, set up the pipeline and train the tagger. In order to
    train the tagger with a custom tag map, we're creating a new Language
    instance with a custom vocab.
    """
    args = get_args()
    print(args)
    curr_dir = path.dirname(path.abspath(__file__))
    lang = args.language
    print(lang)
    output_dir = path.join(path.dirname(path.abspath(__file__)), "outputs")
    if not os.path.exists(output_dir):
            os.makedirs(output_dir)

    model_path = "%s/spacymodels/%s/%s.model" % (curr_dir, args.language,  args.tag_type)    
    data_path = "%s/data/train/%s/train.%s.conll" % (curr_dir, args.language, args.encoding)

    file = open(data_path, "r")
    TRAIN_DATA= spacy_data_reader.spacy_load_data(data_path)

    nlp = spacy.blank(lang)
    # add the tagger to the pipeline
    # nlp.create_pipe works for built-ins that are registered with spaCy
    tagger = nlp.create_pipe('tagger')
    # Add the tags. This needs to be done before you start training.
    for tag, values in TAG_MAP.items():
        tagger.add_label(tag, values)
    nlp.add_pipe(tagger)

    optimizer = nlp.begin_training()
    for i in range(n_iter):
        random.shuffle(TRAIN_DATA)
        losses = {}
        # batch up the examples using spaCy's minibatch
        batches = minibatch(TRAIN_DATA, size=compounding(4., 32., 1.001))
        for batch in batches:
            texts, annotations = zip(*batch)
            nlp.update(texts, annotations, sgd=optimizer, losses=losses)
        print('Losses', losses)

    # test the trained model
    test_text = "నా నా కధ అందరి అందరి ఆడపిల్లల కధే ."
    doc = nlp(test_text)
    print('Tags', [(t.text, t.tag_, t.pos_) for t in doc])

    # save model to output directory
    if output_dir is not None:
        output_dir = Path(output_dir)
        if not output_dir.exists():
            output_dir.mkdir()
        nlp.to_disk(output_dir)
        print("Saved model to", output_dir)

        # test the save model
        print("Loading from", output_dir)
        nlp2 = spacy.load(output_dir)
        doc = nlp2(test_text)
        print('Tags', [(t.text, t.tag_, t.pos_) for t in doc])

if __name__ == '__main__':
    pipeline()



from sklearn.metrics import classification_report, confusion_matrix
from sklearn.preprocessing import LabelBinarizer
from itertools import chain

def bio_classification_report(y_true, y_pred):
    """
    Classification report for a list of BIO-encoded sequences.
    It computes token-level metrics and discards "O" labels.
    
    Note that it requires scikit-learn 0.15+ (or a version from github master)
    to calculate averages properly!
    """
    lb = LabelBinarizer()
    y_true_combined = lb.fit_transform(list(chain.from_iterable(y_true)))
    y_pred_combined = lb.transform(list(chain.from_iterable(y_pred)))
        
    tagset = set(lb.classes_) - {'O'}
    tagset = sorted(tagset, key=lambda tag: tag.split('-', 1)[::-1])
    class_indices = {cls: idx for idx, cls in enumerate(lb.classes_)}
    
    return classification_report(
        y_true_combined,
        y_pred_combined,
        labels = [class_indices[cls] for cls in tagset],
        target_names = tagset,
    )

def pos_report(y_true, y_pred):
    """
    Classification report for a list of pos tagged sequences.
    """

    lb = LabelBinarizer()
    y_true_combined = lb.fit_transform(list(chain.from_iterable(y_true)))
    y_pred_combined = lb.transform(list(chain.from_iterable(y_pred)))

    tagset = set(lb.classes_)
    tagset = sorted(tagset, key=lambda tag: tag.split('-', 1)[::-1])
    class_indices = {cls: idx for idx, cls in enumerate(lb.classes_)}

    return classification_report(
        y_true_combined,
        y_pred_combined,
        labels = [class_indices[cls] for cls in tagset],
        target_names = tagset,
    )


"""
Wrapper class.
"""
from seqeval.metrics import f1_score

from hindiwsd.models import BiLSTMCRF, save_model, load_model
from hindiwsd.preprocessing import IndexTransformer
from hindiwsd.tagger import Tagger
from hindiwsd.trainer import Trainer
from hindiwsd.utils import filter_embeddings


class Sequence(object):

    def __init__(self,
                 word_embedding_dim=100,
                 char_embedding_dim=25,
                 word_lstm_size=100,
                 char_lstm_size=25,
                 fc_dim=100,
                 dropout=0.5,
                 embeddings=None,
                 use_char=True,
                 use_crf=True,
                 initial_vocab=None,
                 optimizer='adam'):

        self.model = None
        self.p = None
        self.tagger = None

        self.word_embedding_dim = word_embedding_dim
        self.char_embedding_dim = char_embedding_dim
        self.word_lstm_size = word_lstm_size
        self.char_lstm_size = char_lstm_size
        self.fc_dim = fc_dim
        self.dropout = dropout
        self.embeddings = embeddings
        self.use_char = use_char
        self.use_crf = use_crf
        self.initial_vocab = initial_vocab
        self.optimizer = optimizer

    def fit(self, x_train, y_train, x_valid=None, y_valid=None,
            epochs=1, batch_size=32, verbose=1, callbacks=None, shuffle=True):
        """Fit the model for a fixed number of epochs.

        Args:
            x_train: list of training data.
            y_train: list of training target (label) data.
            x_valid: list of validation data.
            y_valid: list of validation target (label) data.
            batch_size: Integer.
                Number of samples per gradient update.
                If unspecified, `batch_size` will default to 32.
            epochs: Integer. Number of epochs to train the model.
            verbose: Integer. 0, 1, or 2. Verbosity mode.
                0 = silent, 1 = progress bar, 2 = one line per epoch.
            callbacks: List of `keras.callbacks.Callback` instances.
                List of callbacks to apply during training.
            shuffle: Boolean (whether to shuffle the training data
                before each epoch). `shuffle` will default to True.
        """
        p = IndexTransformer(initial_vocab=self.initial_vocab, use_char=self.use_char)
        p.fit(x_train, y_train)
        embeddings = filter_embeddings(self.embeddings, p._word_vocab.vocab, self.word_embedding_dim)

        model = BiLSTMCRF(char_vocab_size=p.char_vocab_size,
                          word_vocab_size=p.word_vocab_size,
                          num_labels=p.label_size,
                          word_embedding_dim=self.word_embedding_dim,
                          char_embedding_dim=self.char_embedding_dim,
                          word_lstm_size=self.word_lstm_size,
                          char_lstm_size=self.char_lstm_size,
                          fc_dim=self.fc_dim,
                          dropout=self.dropout,
                          embeddings=embeddings,
                          use_char=self.use_char,
                          use_crf=self.use_crf)
        model, loss = model.build()
        model.compile(loss=loss, optimizer=self.optimizer)

        trainer = Trainer(model, preprocessor=p)
        trainer.train(x_train, y_train, x_valid, y_valid,
                      epochs=epochs, batch_size=batch_size,
                      verbose=verbose, callbacks=callbacks,
                      shuffle=shuffle)

        self.p = p
        self.model = model

    def predict(self, x_test):
        """Returns the prediction of the model on the given test data.

        Args:
            x_test : array-like, shape = (n_samples, sent_length)
            Test samples.

        Returns:
            y_pred : array-like, shape = (n_smaples, sent_length)
            Prediction labels for x.
        """
        if self.model:
            lengths = map(len, x_test)
            x_test = self.p.transform(x_test)
            y_pred = self.model.predict(x_test)
            y_pred = self.p.inverse_transform(y_pred, lengths)
            return y_pred
        else:
            raise OSError('Could not find a model. Call load(dir_path).')

    def score(self, x_test, y_test):
        """Returns the f1-micro score on the given test data and labels.

        Args:
            x_test : array-like, shape = (n_samples, sent_length)
            Test samples.

            y_test : array-like, shape = (n_samples, sent_length)
            True labels for x.

        Returns:
            score : float, f1-micro score.
        """
        if self.model:
            x_test = self.p.transform(x_test)
            lengths = map(len, y_test)
            y_pred = self.model.predict(x_test)
            y_pred = self.p.inverse_transform(y_pred, lengths)
            score = f1_score(y_test, y_pred)
            return score
        else:
            raise OSError('Could not find a model. Call load(dir_path).')

    def analyze(self, text, tokenizer=str.split):
        """Analyze text and return pretty format.

        Args:
            text: string, the input text.
            tokenizer: Tokenize input sentence. Default tokenizer is `str.split`.

        Returns:
            res: dict.
        """
        if not self.tagger:
            self.tagger = Tagger(self.model,
                                 preprocessor=self.p,
                                 tokenizer=tokenizer)

        return self.tagger.analyze(text)

    def save(self, weights_file, params_file, preprocessor_file):
        self.p.save(preprocessor_file)
        save_model(self.model, weights_file, params_file)

    @classmethod
    def load(cls, weights_file, params_file, preprocessor_file):
        self = cls()
        self.p = IndexTransformer.load(preprocessor_file)
        self.model = load_model(weights_file, params_file)

        return self



from setuptools import setup
import setuptools

require_packages = ['absl-py==0.15.0',
'astunparse==1.6.3',
'autopep8==1.6.0',
'backports.functools-lru-cache==1.6.4',
'bleach==4.1.0',
'build==0.7.0',
'cachetools==4.2.4',
'certifi==2021.10.8',
'charset-normalizer==2.0.7',
'clang==5.0',
'click==7.1.2',
'colorama==0.4.4',
'docutils==0.17.1',
'flatbuffers==1.12',
'gast==0.4.0',
'google-auth==2.3.0',
'google-auth-oauthlib==0.4.6',
'google-pasta==0.2.0',
'grpcio==1.41.0',
'h5py==3.1.0',
'idna==3.3',
'importlib-metadata==4.8.1',
'indic-transliteration==2.3.2',
'joblib==1.1.0',
'keras==2.6.0',
'Keras-Preprocessing==1.1.2',
'keyring==23.2.1',
'Markdown==3.3.4',
'nltk==3.4.5',
'numpy==1.19.5',
'oauthlib==3.1.1',
'opt-einsum==3.3.0',
'packaging==21.0',
'pandas==1.3.4',
'pbr==2.1.0',
'pep517==0.12.0',
'pkginfo==1.7.1',
'polyglot-tokenizer==2.0.2',
'protobuf==3.19.0',
'pyasn1==0.4.8',
'pyasn1-modules==0.2.8',
'pycodestyle==2.8.0',
'Pygments==2.10.0',
'pyiwn==0.0.5',
'pyparsing==3.0.1',
'python-crfsuite==0.9.7',
'python-dateutil==2.8.2',
'pytz==2021.3',
'pywin32-ctypes==0.2.0',
'readme-renderer==30.0',
'regex==2020.10.11',
'requests==2.26.0',
'requests-oauthlib==1.3.0',
'requests-toolbelt==0.9.1',
'rfc3986==1.5.0',
'rsa==4.7.2',
'scikit-learn==1.0',
'scipy==1.7.1',
'selenium==3.141.0',
'seqeval==1.2.2',
'six==1.15.0',
'spello==1.2.0',
'tensorboard==2.7.0',
'tensorboard-data-server==0.6.1',
'tensorboard-plugin-wit==1.8.0',
'tensorflow==2.6.0',
'tensorflow-estimator==2.6.0',
'termcolor==1.1.0',
'threadpoolctl==3.0.0',
'toml==0.10.2',
'tomli==1.2.1',
'tqdm==4.62.3',
'twine==3.4.2',
'typer==0.3.2',
'typing-extensions==3.7.4.3',
'urllib3==1.26.7',
'webencodings==0.5.1',
'Werkzeug==2.0.2',
'wrapt==1.12.1',
'zipp==3.6.0']

setup(
    name = 'hindiwsd', 
    version = '1.0.9',
    author = 'Praatibh Surana, Mirza Yusuf',
    author_email = 'praatibhsurana@gmail.com', 
    description = 'A pipeline for transliteration of hinglish code mixed data to hindi along with spell correction and word sense disambiguation of hindi words.', 
    url = 'https://github.com/praatibhsurana/Hinglish_Hindi_WSD',
    #packages=['callbacks', 'convert_encoding', 'convert_ssf_conll', 'crf_chunk_features', 'crf_pos_features', 'CRF', 'data_reader', 'evaluate', 'generate_features', 'hindi_wsd', 'hipos', 'layers', 'lesks', 'models', 'pipeline', 'preprocessing', 'spacy_data_reader', 'spacy_tagger_test', 'spacy_tagger', 'tagger', 'trainer', 'utils', 'wrapper', 'writer', 'wsd'],
    packages = setuptools.find_packages(where="src"),
    install_requires=require_packages,
    package_data={'hindiwsd':['*.model', '*.pkl', '*.txt', '*.tsv']},
    include_package_data=True,
    package_dir={"":"src"},
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    keywords='nlp, machine learning, spell correction, IndoWordNet, Hindi, Hinglish code mixed to Hindi transliteration, word sense disambiguation',
)

# print(setuptools.find_packages(include=['hindi_wsd.']))


import os
import codecs
import logging


def write_to_file(text, filename):
    with codecs.open(filename, 'w', encoding='utf8', errors='ignore') as fp:
        fp.write(text)


def write_anno_to_file(filename, X_data, y_data, tag_type):
    with codecs.open(filename, 'w', encoding='utf8', errors='ignore') as fp:
        text = ""
        # print(y_data)
        # print(y_data.shape)
        for i, X_sent in enumerate(X_data):
            # fp.write("<s>\n")
            fp.write("\n")
            for j, X_token in enumerate(X_sent):
                if tag_type == "pos":
                    X_token[1] = y_data[i][j]
                if tag_type == "chunk":
                    X_token[2] = y_data[i][j]
                if tag_type == "ner":
                    X_token[1] = y_data[i][j]
                fp.write("%s\n" % "\t".join(X_token))
                if "।" in "\t".join(X_token):
                    fp.write("\n")
                    # fp.write("</s>\n")
                    # fp.write("<s>\n")
                if "." in "\t".join(X_token):
                    fp.write("\n")
                    # fp.write("</s>\n")
                    # fp.write("<s>\n")
            # fp.write("</s>\n")
            fp.write("\n")


def write_to_screen(filename):
    with codecs.open(filename, "r", encoding='utf8', errors='ignore') as fp:
        text = fp.read()
        return text


def mkdirp(path):
    if path == '':
        return
    try:
        os.makedirs(path)
    except OSError as exc:  # Python >2.5
        if exc.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            raise


def set_logger(method, out_dir=None):
    console_format = '[%(levelname)s] (%(name)s) %(message)s'
    logger = logging.getLogger()
    logger.setLevel(logging.DEBUG)
    console = logging.StreamHandler()
    console.setLevel(logging.DEBUG)
    console.setFormatter(logging.Formatter(console_format))
    logger.addHandler(console)
    if out_dir:
        file_format = '[%(levelname)s] (%(name)s) %(message)s'
        log_file = logging.FileHandler(
            out_dir + '/%s.txt' % (method), mode='w')
        log_file.setLevel(logging.DEBUG)
        log_file.setFormatter(logging.Formatter(file_format))
        logger.addHandler(log_file)



# -*- coding: utf-8 -*-
"""
Preprocessors.
"""
import re

import numpy as np
#from allennlp.modules.elmo import Elmo, batch_to_ids
from sklearn.base import BaseEstimator, TransformerMixin
import joblib
from tensorflow.keras.utils import to_categorical
from tensorflow.keras.preprocessing.sequence import pad_sequences

from hindiwsd.utils import Vocabulary


def normalize_number(text):
    return re.sub(r'[0-9０１２３４５６７８９]', r'0', text)


class IndexTransformer(BaseEstimator, TransformerMixin):
    """Convert a collection of raw documents to a document id matrix.

    Attributes:
        _use_char: boolean. Whether to use char feature.
        _num_norm: boolean. Whether to normalize text.
        _word_vocab: dict. A mapping of words to feature indices.
        _char_vocab: dict. A mapping of chars to feature indices.
        _label_vocab: dict. A mapping of labels to feature indices.
    """

    def __init__(self, lower=True, num_norm=True,
                 use_char=True, initial_vocab=None):
        """Create a preprocessor object.

        Args:
            lower: boolean. Whether to convert the texts to lowercase.
            use_char: boolean. Whether to use char feature.
            num_norm: boolean. Whether to normalize text.
            initial_vocab: Iterable. Initial vocabulary for expanding word_vocab.
        """
        self._num_norm = num_norm
        self._use_char = use_char
        self._word_vocab = Vocabulary(lower=lower)
        self._char_vocab = Vocabulary(lower=False)
        self._label_vocab = Vocabulary(lower=False, unk_token=False)

        if initial_vocab:
            self._word_vocab.add_documents([initial_vocab])
            self._char_vocab.add_documents(initial_vocab)

    def fit(self, X, y):
        """Learn vocabulary from training set.

        Args:
            X : iterable. An iterable which yields either str, unicode or file objects.

        Returns:
            self : IndexTransformer.
        """
        self._word_vocab.add_documents(X)
        self._label_vocab.add_documents(y)
        if self._use_char:
            for doc in X:
                self._char_vocab.add_documents(doc)

        self._word_vocab.build()
        self._char_vocab.build()
        self._label_vocab.build()

        return self

    def transform(self, X, y=None):
        """Transform documents to document ids.

        Uses the vocabulary learned by fit.

        Args:
            X : iterable
            an iterable which yields either str, unicode or file objects.
            y : iterabl, label strings.

        Returns:
            features: document id matrix.
            y: label id matrix.
        """
        word_ids = [self._word_vocab.doc2id(doc) for doc in X]
        word_ids = pad_sequences(word_ids, padding='post')

        if self._use_char:
            char_ids = [[self._char_vocab.doc2id(w) for w in doc] for doc in X]
            char_ids = pad_nested_sequences(char_ids)
            features = [word_ids, char_ids]
        else:
            features = word_ids

        if y is not None:
            y = [self._label_vocab.doc2id(doc) for doc in y]
            y = pad_sequences(y, padding='post')
            y = to_categorical(y, self.label_size).astype(int)
            # In 2018/06/01, to_categorical is a bit strange.
            # >>> to_categorical([[1,3]], num_classes=4).shape
            # (1, 2, 4)
            # >>> to_categorical([[1]], num_classes=4).shape
            # (1, 4)
            # So, I expand dimensions when len(y.shape) == 2.
            y = y if len(y.shape) == 3 else np.expand_dims(y, axis=0)
            return features, y
        else:
            return features

    def fit_transform(self, X, y=None, **params):
        """Learn vocabulary and return document id matrix.

        This is equivalent to fit followed by transform.

        Args:
            X : iterable
            an iterable which yields either str, unicode or file objects.

        Returns:
            list : document id matrix.
            list: label id matrix.
        """
        return self.fit(X, y).transform(X, y)

    def inverse_transform(self, y, lengths=None):
        """Return label strings.

        Args:
            y: label id matrix.
            lengths: sentences length.

        Returns:
            list: list of list of strings.
        """
        y = np.argmax(y, -1)
        inverse_y = [self._label_vocab.id2doc(ids) for ids in y]
        if lengths is not None:
            inverse_y = [iy[:l] for iy, l in zip(inverse_y, lengths)]

        return inverse_y

    @property
    def word_vocab_size(self):
        return len(self._word_vocab)

    @property
    def char_vocab_size(self):
        return len(self._char_vocab)

    @property
    def label_size(self):
        return len(self._label_vocab)

    def save(self, file_path):
        joblib.dump(self, file_path)

    @classmethod
    def load(cls, file_path):
        p = joblib.load(file_path)

        return p


def pad_nested_sequences(sequences, dtype='int32'):
    """Pads nested sequences to the same length.

    This function transforms a list of list sequences
    into a 3D Numpy array of shape `(num_samples, max_sent_len, max_word_len)`.

    Args:
        sequences: List of lists of lists.
        dtype: Type of the output sequences.

    # Returns
        x: Numpy array.
    """
    max_sent_len = 0
    max_word_len = 0
    for sent in sequences:
        max_sent_len = max(len(sent), max_sent_len)
        for word in sent:
            max_word_len = max(len(word), max_word_len)

    x = np.zeros((len(sequences), max_sent_len, max_word_len)).astype(dtype)
    for i, sent in enumerate(sequences):
        for j, word in enumerate(sent):
            x[i, j, :len(word)] = word

    return x



import sys,os.path as path
sys.path.append(path.dirname(path.dirname(path.dirname(path.abspath(__file__)))))

#from hindiwsd.
from crf_pos_features import crf_pos_features
#from hindiwsd.
from crf_chunk_features import crf_chunk_features
import numpy as np

def sent2features(sent, tag_type, model_type):
    if tag_type == "pos":
    	if model_type == "crf":
        	return [crf_pos_features(sent, i) for i in range(len(sent))]
    if tag_type == "chunk":
    	if model_type == "crf":
        	return [crf_chunk_features(sent, i) for i in range(len(sent))]

def sent2labels(sent, tag_type):
    if tag_type == "pos":
        return [postag for token, postag, chunk in sent]
    if tag_type == "chunk":
        return [chunk for token, postag, chunk in sent]

def sent2tokens(sent):
    return [token for token, postag, chunk in sent]


def append_tags(sents, tag_type, pred):
	#To do make it more efficient
	if tag_type == "pos":
		for i, sent in enumerate(sents):
			for j, vals in enumerate(sent): 
				if tag_type == "pos":
					sents[i][j][1] = pred[i][j]
				if tag_type == "chunk":
					sents[i][j][2] = pred[i][j]
	return sents

	"""

	np_sents = np.array(sents)
	if tag_type == "pos":
		np_sents[:,:,1] = np.array(pred)
	if tag_type == "chunk":
		np_sents[:,:,2] =  np.array(pred)

	return np_sents
	"""


import re, os
import codecs
import logging
from irtokz import IndicTokenizer

logger = logging.getLogger(__name__)

def spacy_load_data(filename):
    file = open(filename, "r")
    sentence = ""
    posList = list()
    wordList = list()
    fullList = list()
    for line in file.readlines():
        if (line.strip() == ""):
            dict = {}
            dict['tags'] = posList
            dict['words'] = wordList
            tuple1 = (sentence.strip(), dict)
            fullList.append(tuple1)
            sentence = ""
            posList = list()
            wordList = list()
        else:
            array = re.split('\s+', line.strip())
            if (array[1] != "" and array[2] != ""):
                sentence = sentence + array[1] + " "
                posList.append(array[2])
                wordList.append(array[1])
    return fullList
